#!/usr/bin/perl -w

###########################################################################
# This script solves Reflexion puzzles.
###########################################################################

use IPC::Open2;

$PROG = "reflex";
$WALL = -1; $DIAMOND = -2; $START = -3; $FINISH = -4; $SPACE = -5;

########### process command-line options ###########

sub usage {
	print<<EOT;

usage: $PROG [<options>] [<filename>] -- solve Reflexion puzzles
	<filename> : contains formula; if none given, reads from stdin

	Options:
		-d: just display the puzzle (respects -m and -t)
		-g: generate and show formula, do not pass to solver
		-h: print this help message
		-m: Metapost output
		-q: quiet (reduce messages)
		-r: do not generate ranks
		-s: generate SAT formula instead of SMT (implies -r)
		-t: transpose the puzzle
		-z: show response of solver, do not interpret

	Note: This script requires the Z3 solver (z3) to be in your path.
	The Metapost-to-PDF compiler (mptopdf) is required when using -m.

EOT
	exit 2;
}

$options = {
	'b' => \$opt_blank,		# inutile
	'd' => \$opt_display,
	'g' => \$opt_showformula,
	'h' => \$opt_help,
	'm' => \$opt_mpost,
	'q' => \$opt_quiet,
	'r' => \$opt_norank,
	's' => \$opt_sat,
	't' => \$opt_transpose,
	'z' => \$opt_showresponse
};

for $v (values $options) { $$v = 0; }
$fn = undef;
for $arg (@ARGV) {
	if (substr($arg,0,1) eq "-") {
		for $a (split(//,substr($arg,1))) {
			die "unknown option -$a" unless defined($options->{$a});
			${$options->{$a}} = 1;
		}
	} else {
		usage() if defined($fn);
		$fn = $arg;
	}
}

usage() if $opt_help;
$opt_smt = !$opt_sat;
$opt_norank = 1 unless $opt_smt;
$opt_norank = 1 if $opt_display;

print STDERR "$PROG -- (invoke with -h for help)\n" unless $opt_quiet;
if (defined($fn)) {
	die "cannot read file $fn" unless -r $fn;
	open STDIN,$fn;
	print STDERR "reading formula from file $fn\n" unless $opt_quiet;
} else {
	print STDERR "reading formula from standard input\n" unless $opt_quiet;
}

########### read puzzle and check for conformity ###########

$mirrors = $diams = $spaces = 0;
$width = 0;
$height = 0;
$isfullwall = 0;
$sx = $sy = $fx = $fy = -1;
$starts = $finishes = 0;
$fields = 0;

while (<STDIN>) {
	tr/\015\012//d;
	$line = $_;
	if ($width > 0 && length($line) != $width) {
		die "not a rectangle\n";
	}
	$width = length($line) unless $width > 0;

	$isfullwall = ($line =~ /^[SE%]*$/);
	die "upper wall not closed" if $height == 0 && !$isfullwall;
	die "side walls not closed" if $line !~ /^[SE%].*[SE%]$/;

	for ($i = 0; $i < $width; $i++) {
		$c = substr($line,$i,1);
		if ($c eq '%') { $pz[$i][$height] = $WALL; }
		elsif ($c eq 'D') { $pz[$i][$height] = $DIAMOND; $diams++; }
		elsif ($c eq 'M') { $pz[$i][$height] = ++$mirrors; }
		elsif ($c eq '/') { $pz[$i][$height] = ++$mirrors;
				    push @fixed,$mirrors; }
		elsif ($c eq '\\') { $pz[$i][$height] = ++$mirrors;
				    push @fixed,-$mirrors; }
		elsif ($c eq 'S') { $sx = $i; $sy = $height; $starts++;
				    $pz[$i][$height] = $START; }
		elsif ($c eq 'E') { $fx = $i; $fy = $height; $finishes++;
				    $pz[$i][$height] = $FINISH; }
		elsif ($c eq ' ') { $pz[$i][$height] = $SPACE; $spaces++}
		else { die "unknown content"; }

		# give an identifier to the field if necessary
		if ($c =~ /[\/\\DM]/ || ($opt_blank && $c eq ' '))
			{ $fd[$i][$height] = ++$fields; }
		$fd[$i][$height] = -1 if $c =~ /[SE%]/;
	}
	$height++;
}

die "empty puzzle" if $width <= 2 || $height <= 2;
die "lower wall not closed" if !$isfullwall;
die "multiple start positions" if $starts > 1;
die "no start position" if $starts == 0;
die "multiple end positions" if $finishes > 1;
die "no end position" if $finishes == 0;
die "bad start position" unless $sx == 0 || $sx == $width-1
					|| $sy == 0 || $sy == $height-1;
die "bad end position" unless $fx == 0 || $fx == $width-1 ||
					 $fy == 0 || $fy == $height-1;
die "start position in corner" if ($sx == 0 || $sx == $width-1)
					&& ($sy == 0 || $sy == $height-1);

# make sure start is taken into account
($x,$y) = ($sx,$sy);
if ($x == 0) { $x++; } elsif ($x == $width-1) { $x--; }
elsif ($y == 0) { $y++; } elsif ($y == $height-1) { $y--; }
$fd[$x][$y] = ++$fields unless defined($fd[$x][$y]);

# transpose puzzle (optional)
transpose() if $opt_transpose;

if ($opt_display) {
	$opt_mpost? output_metapost() : output_text();
	exit 0;
}

########### generate the SAT formula ###########

printf STDERR "puzzle alright, generating %s formula...\n\n",
		($opt_smt? "SMT" : "SAT") unless $opt_quiet;

if ($opt_showformula) {
	open Writer,"|cat";
} elsif ($opt_smt) {
	open2(*Reader,*Writer,"z3 -smt2 -in");
} else {
	open2(*Reader,*Writer,"z3 -dimacs -in");
}

header();
for ($i = 1; $i < $width-1; $i++) {
	for ($j = 1; $j < $height-1; $j++) {
		gen($i,$j);
	}
}
genfixed();
footer();

close Writer;
exit 0 if $opt_showformula;

########### interpret response from SAT solver ###########

if ($opt_showresponse) {
	while (<Reader>) { print; }
	exit 0;
}

while (<Reader>) {
	tr/\015\012//d;
	if (/^unsat$/) {
		print "Unfortunately, there is NO SOLUTION.\n";
		exit 1;
	}
	if (/^sat$/) {
		print "There is a solution as follows:\n" unless $opt_quiet;
		next;
	}
        ($dir,$j) = ($1,$2) if /define-fun r(.)_(\d*)/;
	$v{$dir.$j} = $1 if /(\d+)\)/;
        $i = $1 if /define-fun v_(\d*)/;
	$v[$i] = $i if /true/;
	$v[$i] = -$i if /false/;
	next if /[()]/;
	@a = split;
	for $l (@a) {
		$i = ($l > 0)? $l : -$l;
		$v[$i] = $l;
	}
}

$opt_mpost? output_metapost() : output_text();

exit 0;

######## generate formula for a field (only inside external walls) ########

sub gen {
	my ($i,$j) = @_;

	# content of this field
	my $c = $pz[$i][$j];
	my $f = $fd[$i][$j];

	# if it's a wall, nothing do do; surrounding fields will take care
	return if $c eq $WALL;

	# skip blank spaces
	return unless defined($f);

	# if it's a diamond, require the field to be entered
	gentouch($i,$j) if $c eq $DIAMOND;

	# conditions for entering the position
	genentries($i,$j,$f);

	# conditions for leaving the position
	genstraight($i,$j,$f) if $c eq $DIAMOND || $c eq $SPACE;
	genmirror($i,$j,$f) if $c >= 0;
}

sub genentries {
	my ($i,$j,$f) = @_;

	genentry($i,$j,$f,'n','s',0,-1);
	genentry($i,$j,$f,'e','w',1,0);
	genentry($i,$j,$f,'s','n',0,1);
	genentry($i,$j,$f,'w','e',-1,0);
}

sub genentry {
	my ($i,$j,$f,$dir,$opd,$dx,$dy) = @_;

	# find the relevant neighbour
	my ($vi,$vj) = ($i,$j);
	do { $vi += $dx; $vj += $dy; } while (!defined($fd[$vi][$vj]));
	my $vs = $pz[$vi][$vj];

	# walls must not be touched, ball must come from S and leave towards E
	if ($vs eq $WALL)	{ clause(-entry($i,$j,$dir));
				  clause(-sortie($i,$j,$dir)); }
	elsif ($vs eq $START)	{ clause(entry($i,$j,$dir));
				  clause(-sortie($i,$j,$dir)); }
	elsif ($vs eq $FINISH)	{ clause(-entry($i,$j,$dir));
				  clause(sortie($i,$j,$dir)); }

	# otherwise it enters this position if it comes from neighbour
	else {
		eqclause(sortie($vi,$vj,$opd),entry($i,$j,$dir));

		return if $opt_norank;
		my $vf = $fd[$vi][$vj];
		my $e = entry($i,$j,$dir);
		print Writer "(assert (=> v_$e (< r${opd}_$vf r${dir}_$f)))\n";
		print Writer "(assert (< 0 r${opd}_$vf))\n";
	}
}

# field is entered from any direction
sub gentouch {
	my ($i,$j) = @_;
	clause(entry($i,$j,'n'),entry($i,$j,'e'),
		entry($i,$j,'s'),entry($i,$j,'w'));
}

# ball travels straight through empty space
sub genstraight {
	my ($i,$j,$f) = @_;
	eqclause(entry($i,$j,'n'),sortie($i,$j,'s'));
	eqclause(entry($i,$j,'e'),sortie($i,$j,'w'));
	eqclause(entry($i,$j,'s'),sortie($i,$j,'n'));
	eqclause(entry($i,$j,'w'),sortie($i,$j,'e'));

	return if $opt_norank;
	print Writer "(assert (= rn_$f rs_$f))\n";
	print Writer "(assert (= rw_$f re_$f))\n";
}

# conditions for mirror; mirror variable true = '/', false = '\'
sub genmirror {
	my ($i,$j,$f) = @_;
	my $m = mivar($pz[$i][$j]);
	genref($i,$j,$m,'n','w','e');
	genref($i,$j,$m,'e','s','n');
	genref($i,$j,$m,'s','e','w');
	genref($i,$j,$m,'w','n','s');

	return if $opt_norank;
	print Writer "(assert (=> v_$m (= rn_$f rw_$f)))\n";
	print Writer "(assert (=> v_$m (= rs_$f re_$f)))\n";
	print Writer "(assert (=> (not v_$m) (= rn_$f re_$f)))\n";
	print Writer "(assert (=> (not v_$m) (= rs_$f rw_$f)))\n";
}

sub genref {
	my ($i,$j,$m,$entry,$left,$right) = @_;
	# entry and mirror => exit
	clause(-entry($i,$j,$entry),-$m,sortie($i,$j,$left));
	clause(-entry($i,$j,$entry),$m,sortie($i,$j,$right));
	# exit => mirror and entry
	clause(-sortie($i,$j,$entry),-$m,entry($i,$j,$left));
	clause(-sortie($i,$j,$entry),$m,entry($i,$j,$right));
}

# conditions for mirrors with fixed orientation
sub genfixed {
	for my $m (@fixed) {
		clause(($m > 0)? mivar($m) : -mivar(-$m));
	}
}

######## helper procedures ########

sub header {
	my $nv = $fields*8 + $mirrors;
	my $nc = $diams*17 + $mirrors*24 + $#fixed+1;
	$nc += $spaces*16 if $opt_blank;

	if ($opt_smt) {
		for ($i = 1; $i <= $nv; $i++) {
			print Writer "(declare-const v_$i Bool)\n";
		}
		return if $opt_norank;
		for ($i = 1; $i <= $fields; $i++) {
			print Writer "(declare-const rn_$i Int)\n";
			print Writer "(declare-const re_$i Int)\n";
			print Writer "(declare-const rs_$i Int)\n";
			print Writer "(declare-const rw_$i Int)\n";
		}
	} else {
		print Writer "p cnf $nv $nc\n";
	}
}

sub footer {
	print Writer "(check-sat)\n(get-model)\n" if $opt_smt;
}

# output a single clause
sub clause {
	if ($opt_smt) {
		print Writer "(assert (or ", join(" ",
			map { $_ > 0? "v_$_" : "(not v_".(-$_).")" } @_),
			"))\n";
	} else {
		print Writer join(" ",@_)," 0\n";
	}
}

# generate x <-> y constraint
sub eqclause {
	my ($x,$y) = @_;
	clause(-$x,$y);
	clause(-$y,$x);
}

# get index for "ball enters position i,j from direction d"
sub entry {
	my ($i,$j,$d) = @_;
	my $v = $fd[$i][$j]*8;

	return $v-7 if $d eq 'n';
	return $v-6 if $d eq 'e';
	return $v-5 if $d eq 's';
	return $v-4 if $d eq 'w';
}

# get index for "ball leaves position i,j in direction d"
sub sortie {
	return entry(@_)+4;
}

# get index for a mirror variable
sub mivar {
	my ($id) = @_;
	return $fields*8+$id;
}

######## transformations ########

sub transpose {
	my (@newpz, @newfd);
	for ($j = 0; $j < $height; $j++) {
		for ($i = 0; $i < $width; $i++) {
			$newpz[$j][$i]  = $pz[$i][$j];
			$newfd[$j][$i]  = $fd[$i][$j];
		}
	}
	@pz = @newpz;
	@fd = @newfd;
	($sx,$sy,$fx,$fy,$width,$height) = ($sy,$sx,$fy,$fx,$height,$width);
}

######## output procedures ########

sub output_text {
	for ($j = 0; $j < $height; $j++) {
		for ($i = 0; $i < $width; $i++) {
			$c = $pz[$i][$j];
			if ($c eq $WALL) { print '%'; }
			elsif ($c eq $START) { print 'S'; }
			elsif ($c eq $FINISH) { print 'E'; }
			elsif ($c eq $DIAMOND) { print 'D'; }
			elsif ($c eq $SPACE) { print ' '; }
			elsif ($c >= 0) {
				$z = $v[mivar($c)];
				print ((defined($z) && $z > 0)? '/' : '\\');
			}
		}
		printf "\n";
	}
}

sub output_metapost {
	open FD,">/tmp/reflex.$$.mp";
	select FD;
	reflexion_mp();
	for ($j = 0; $j < $height; $j++) {
		for ($i = 0; $i < $width; $i++) {
			$c = $pz[$i][$j];
			if ($c eq $WALL) { print "wall($i,$j);\n"; }
			elsif ($c eq $START) { print "start($i,$j);\n"; }
			elsif ($c eq $FINISH) { print "target($i,$j);\n"; }
			elsif ($c eq $DIAMOND) { print "diamond($i,$j);\n"; }
			elsif ($c >= 0) {
				$z = $v[mivar($c)];
				$z = ((defined($z) && $z > 0)? '/' : '\\');
				print "mirror($i,$j,\"$z\");\n";
			}
			getrank($i,$j,"n"); getrank($i,$j,"e");
			getrank($i,$j,"s"); getrank($i,$j,"w");
		}
	}
	print "move_ball(($sx,$sy)--",
		join("--",map { $bp{$_} } sort {$a <=> $b} keys %bp),
		"--($fx,$fy));\n" unless $opt_norank;

	print "draw_all;\nend.\n";
	close FD;
	select STDOUT;
	my $redir = ($opt_quiet? " > /dev/null" : "");
	system "cd /tmp; mptopdf reflex.$$.mp$redir; evince reflex.$$-0.pdf";
	@files = qw/.log .mp .mpx .0 -0.pdf/;
	unlink(map { "/tmp/reflex.$$".$_ } @files);
}

sub getrank {
	my ($i,$j,$dir) = @_;
	return if $opt_norank;
	my $f = $fd[$i][$j];
	return if !defined($f) || $f < 0;
	return if $v[entry($i,$j,$dir)] < 0;
	$bp{$v{$dir.$f}} = "($i,$j)";
}

sub reflexion_mp {
	print<<'EOT';
numeric unit;
unit=5/9cm;

color c_wall,c_grid,c_dshaded,c_mirror,c_diamond,c_path;
c_wall=0.3white;
c_grid=0.93white;
c_dshaded=0.7white;
c_mirror=0.8blue;
c_diamond=0.7red;
c_path=0.7green;

unitsquare:=unitsquare shifted (-1/2*(1,1));

path mirrorpath,diampath,ballpath;
mirrorpath=(-0.4,-0.1)--(0.4,-0.1)--(0.4,0.1)--(-0.4,0.1)--cycle;
diampath=(0.3,0.1)--(0.1,0.3)--(-0.1,0.3)--(-0.3,0.1)--(-0.3,-0.1)--(-0.1,-0.3)--(0.1,-0.3)--(0.3,-0.1)--cycle;

numeric last_wall,last_mirror,last_diamond;
numeric startx, starty, targetx, targety, bpset;
numeric xmin,ymin,xmax,ymax;
numeric wallx[],wally[],mirrorx[],mirrory[],mirrora[],diamondx[],diamondy[];
last_wall=0;
last_mirror=0;
last_diamond=0;
startx:=0;
targetx:=0;
bpset:=0;

def wall (expr xi,yi) =
	i:=incr(last_wall);
	wallx[i]:=xi; wally[i]:=-yi;
enddef;

vardef start(expr x,y) = startx:=x; starty:=-y; enddef;
vardef target(expr x,y) = targetx:=x; targety:=-y; enddef;

def mirror(expr xi,yi,oi) =
	i:=incr(last_mirror);
	mirrorx[i]:=xi;mirrory[i]:=-yi;
	if (oi="/"): mirrora[i]:=45; fi;
	if (oi="\"): mirrora[i]:=315; fi;
enddef;

def diamond(expr xi,yi) =
	i:=incr(last_diamond);
	diamondx[i]:=xi; diamondy[i]:=-yi;
enddef;

def move_ball(expr p) =
	bpset:=1;
	ballpath:=p;
enddef;

vardef draw_grid (expr xi,yi,lx,ly) = 
	for i=0 upto lx:
		draw ((xi+i-1/2,yi-1/2)--(xi+i-1/2,yi+ly-1/2)) scaled unit withcolor c_grid;
	endfor;
	for i=0 upto ly:
		draw ((xi-1/2,yi+i-1/2)--(xi+lx-1/2,yi+i-1/2)) scaled unit withcolor c_grid;
	endfor;
enddef;

vardef draw_all =
	beginfig(0);
	if (bpset=1):
	draw ballpath scaled unit reflectedabout ((0,0),(1,0)) dashed evenly withcolor c_path;
	fi;
	for i=1 upto last_wall:
		fill (unitsquare shifted (wallx[i],wally[i])) scaled unit withcolor c_wall;
	endfor;
	fill (unitsquare shifted (startx,starty)) scaled unit withcolor c_dshaded;
	fill (unitsquare shifted (targetx,targety)) scaled unit withcolor c_dshaded;
	xmin:=wallx[1]; xmax:=wallx[1];
	ymin:=wally[1]; ymax:=wally[1];
	for i=1 upto last_wall:
		if (wallx[i] < xmin): xmin:=wallx[i]; fi;
		if (wally[i] < ymin): ymin:=wally[i]; fi;
		if (wallx[i] > xmax): xmax:=wallx[i]; fi;
		if (wally[i] > ymax): ymax:=wally[i]; fi;
	endfor;
	draw_grid(xmin,ymin,xmax-xmin+1,ymax-ymin+1);
	label(btex $S$ etex scaled 1.5,(startx,starty) scaled unit);
	label(btex $E$ etex scaled 1.5,(targetx,targety) scaled unit);
        for i=1 upto last_mirror:
	    fill mirrorpath rotated mirrora[i] shifted (mirrorx[i],mirrory[i])
			scaled unit withcolor c_mirror;
	    draw mirrorpath rotated mirrora[i] shifted (mirrorx[i],mirrory[i])
			scaled unit;
        endfor;
        for i=1 upto last_diamond:
	    fill diampath shifted (diamondx[i],diamondy[i])
			scaled unit withcolor c_diamond;
	    draw diampath shifted (diamondx[i],diamondy[i]) scaled unit;
        endfor;
	endfig;
enddef;
EOT
}
