#!/usr/bin/perl -w

###########################################################################
# This script is a generalization of 3sat-reflex; it generates a Reflexion
# puzzle from a SAT formula. The formula should be given in DIMACS format.
# There is no restriction on the number of literals per clause.
###########################################################################

$PROG = "sat";

########### process command-line options ###########

sub usage {
	print <<EOT;

$PROG -- generates a Reflexion puzzle from a CNF formula

usage: $PROG [<options>] [<filename>]
		<filename> : contains SAT formula in DIMACS format
		if no filename is given, reads from stdin

	Options:
		-h: print this help message
		-q: quiet (reduce messages)
		-S<opts>: directly invoke solver with given options

EOT
		exit 2;
}

$options = {
	'h' => \$opt_help,
	'q' => \$opt_quiet,
};

for $v (values $options) { $$v = 0; }
$opt_solver = 0;
$fn = undef;
for $arg (@ARGV) {
	if (substr($arg,0,1) eq "-") {
		for ($i = 1; $i < length($arg); $i++) {
			$a = substr($arg,$i,1);
			if ($a eq 'S') {
				$opt_solver = 1;
				$solver_options = substr($arg,$i+1);
				$solver_options =~ s/^=//;
				$arg = substr($arg,0,$i);
				next;
			}
			die "unknown option -$a" unless defined($options->{$a});
			${$options->{$a}} = 1;
		}
	} else {
		usage() if defined($fn);
		$fn = $arg;
	}
}

usage() if $opt_help;
if ($opt_solver) {
	$solver_options .= "q" if $opt_quiet;
	$opt_quiet = 1 if $solver_options =~ /q/;
	$solver_options = "-".$solver_options if $solver_options ne "";
}

print STDERR "$PROG -- (invoke with -h for help)\n" unless $opt_quiet;
if (defined($fn)) {
	die "cannot read file $fn" unless -r $fn;
	open STDIN,$fn;
	print STDERR "reading formula from file $fn\n" unless $opt_quiet;
} else {
        print STDERR "reading formula from standard input\n" unless $opt_quiet;
}

########### read DIMACS file ###########

$max = 0;
$nc = 0;
$ms = 0;

while (<STDIN>) {
	tr/\015\012//d;
	next if /^p cnf/;
	next if /^c/;
	die "bad file format" unless /^(\s*-?\d+)*\s*$/;
	$nc++;
	push @clauses,$_;
	@nums = split;
	die "empty clause" if $#nums <= 0;
	die "clause not 0-terminated" if $nums[$#nums] != 0;
	$ms = $#nums if $#nums > $ms;
	for $i (@nums) {
		$max = $i if $i > $max;
		$max = -$i if -$i > $max;
	}
}

print STDERR "formula alright, generating puzzle...\n\n" unless $opt_quiet;

########### compute size of puzzle ###########

# number of variables and literals
$nv = $max;
$nl = 2*$max;

# width of central bus
$cw = 0;
for (@clauses) {
	@nums = split;
	$cw += 1 if $#nums == 1;
	$cw += 8 if $#nums == 2;
	$cw += 11 if $#nums == 3;
	$cw += 5+6*$#nums if $#nums > 3;
}
$cw = 2 if $cw < 2;

# height of clause boxes
$baserow = $nv*6+1;
$blines = 6 if $ms == 1;
$blines = 8 if $ms == 2;
$blines = 12 if $ms == 3;
$blines = 9+$ms if $ms >= 4;

#total size
$height = 6*$nv+$blines+2;
$width = 4*$nv+$cw+3;

########### generate the puzzle ###########

# puzzle with outer walls
@puzzle = ();
for ($i = 0; $i < $height; $i++) {
	my @line = ('%',(' ')x($width-2),'%',"\n");
	push @puzzle,\@line;
}
put(0,0,'%'x$width);
put(0,$height-1,'%'x$width);
gadgets();

# choices
$pos = $nl+1;
for ($i = 0; $i < $nv; $i++) { put($pos,$i*4+1,('M')x3,'%'); }

# mergers and start
$pos = $nl+$cw;
put($pos,0,'S',' ','M',' ','%');
for ($i = 1; $i < $nv; $i++) { put($pos,$i*4+1,('M')x3,'%'); }

# dividing wall
put($nl+1,4*$nv,'%'x$cw);

# mirror arrays
for ($i = 0; $i < $nl; $i++) {
	put($i+1,2*$i+1,'M');					# upper left
	put($i+1,6*$nv-$i,'M');					# middle left
	put($nl+$cw+2+$i,6*$nv-$i,'M');				# middle right
	put($nl+$cw+2+$i,2*$i+5,'M') unless $i >= $nl-2;	# upper right
}

# bus
for ($i = 1; $i <= $nl; $i++) {
	$v = int(($nl-$i+2)/2);
	$v = -$v if $i % 2;
	$row = 4*$nv+$i;
	$pos = $nl+1;
	
	for (@clauses) {
		@nums = split;
		if ($#nums == 1) {
			put($pos,$row,'D') if $nums[0] == $v;
			$pos += 1;
		} elsif ($#nums == 2) {
			put($pos+1,$row,'M % M') if $nums[0] == $v;
			put($pos+2,$row,'M%M') if $nums[1] == $v;
			$pos += 8;
		} elsif ($#nums == 3) {
			put($pos+1,$row,'M%M') if $nums[0] == $v;
			put($pos+4,$row,'M%M') if $nums[1] == $v;
			put($pos+7,$row,'M%M') if $nums[2] == $v;
			$pos += 11;
		} else {
			$l = $#nums;
			for ($j = 0; $j < $l; $j++) {
				put($pos+3+$j*6,$row,'M % M') if $nums[$j] == $v
			}
			$pos += 5+6*$l;
		}
	}
}

# diamond cases
$pos = $nl+1;
for (@clauses) {
	@nums = split;
	if ($#nums == 1) { dput($pos,0,'%'); $pos += 1; next; }
	elsif ($#nums == 2) { dput($pos,0,@box2); $pos += 7; }
	elsif ($#nums == 3) { dput($pos,0,@box3); $pos += 10; }
	else {
		$l = $#nums;
		dput($pos,0,@box4l);
		dput($pos,5+$l," M","MM",""," M");
		$pos += 2;
		for ($j = 1; $j <= $l; $j++) {
			dput($pos,0,@box4);
			dput($pos,7+$l-$j,"%") unless $j == $l;
			dput($pos,10+$l-$j,"%") unless $j <= 2;
			if ($j == 1) { dput($pos+2,4+$l,@tower1); }
			elsif ($j == 2) { dput($pos+2,3+$l,@tower2); }
			elsif ($j == $l-1) { dput($pos+2,9,reverse(@tower2)); }
			elsif ($j == $l) { dput($pos+2,7,reverse(@tower1)); }
			else { dput($pos+2,5+$l-$j,@cross); }
			$pos += 6;
		}
		dput($pos,0,@box4r); $pos += 2;
	}
	dput($pos,0,('%')x(9+$ms));
	dput($pos,4,' '); $pos++;
}

# additional dividing walls
dput(1,0,'%'x$nl);
dput($pos,0,'%'x($cw+$nl+1-$pos)) if ($pos < $cw+$nl+1);
dput($cw+$nl+1,0,'%'x($nl-1));
dput($nl,1,('%')x$blines);
dput($nl,4,' ');

# final merge and exit
dput(0,4,'E'); dput(-5,3,'MM','M','M M');

########### print puzzle to stdout or solver ###########

if ($opt_solver) {
	printf "passing puzzle to solver...\n" unless $opt_quiet;
	open STDOUT,"| ./reflex.pl $solver_options";
}
for $line (@puzzle) { print join("",@$line); }
close STDOUT if $opt_solver;
exit 0;

########### auxilary procedures ###########

sub dput {
	my ($x,$y,@l) = @_;
	put($x,$baserow+$y,@l);
}

sub put {
	my ($x,$y,@l) = @_;
	for my $s (@l) {
		my $line = $puzzle[$y++];
		my @nc = split(//,$s);
		splice @$line,$x,$#nc+1,@nc;
	}
}

########### various gadgets used during the construction ###########

sub gadgets {
@box2 = split(/\n/,<<EOT);
%  %  %
   %   
   %   
M   D M
M  %  M
MD    M
 M   M 
  M M  
EOT
@box3 = split(/\n/,<<EOT);
% %  %  % 
M       M 
  M  M    
     MD M 
M       M 
     M DM 
MDM       
  M  M    
% %  %  % 
 M M      
    M M   
       M M
EOT
@box4 = split(/\n/,<<EOT);
% %%% 
% M M 
  M%M 
%MD  M
%
% M M 
EOT
@box4l = split(/\n/,<<EOT);
%%
  
 M
  
M 
  
 M
 %
EOT
@box4r = split(/\n/,<<EOT);
%%
  
 M
  
 M
 %
 M
M 
MM
  
M 
EOT
@tower1 = ((" M ")x3," % ","M%M");
@tower2 = ((" M ")x3,"M%M");
@cross = ((" M ")x3,"M%M",(" M ")x3);
}
