#!/usr/bin/perl -w

###########################################################################
# This script takes as input a graph and outputs a Reflexion puzzle that
# has a solution if and only if the graph has a vertex cover of size k.
# The graph must be in the same format as for hamilton-reflex-pl, see there.
# Directed and undirected graphs are treated identically.
# Additionally, k must be specified on the command line (default is 1).
###########################################################################

$PROG = "vertexcover";

########### process command-line options ###########

sub usage {
	print <<EOT;

$PROG -- generates a Reflexion puzzle that searches for a k-vertex cover

usage: $PROG [<options>] [<filename>]
		<filename> : contains directed graph
		if no filename is given, reads from stdin

	Options:
		-d: generate extra diamonds (helps the solver sometimes)
		-h: print this help message
		-k=<int>: set value of k (default 1)
		-q: quiet (reduce messages)
		-S<opts>: directly invoke solver with given options

EOT
		exit 2;
}

$options = {
	'd' => \$opt_diam,
	'h' => \$opt_help,
	'q' => \$opt_quiet,
	'x' => \$opt_altx		# inutile
};

for $v (values $options) { $$v = 0; }
$opt_solver = 0;
$fn = undef;
$k = 1;
for $arg (@ARGV) {
	if (substr($arg,0,1) eq "-") {
		for ($i = 1; $i < length($arg); $i++) {
			$a = substr($arg,$i,1);
			if ($a eq 'k') {
				$b = substr($arg,$i+1);
				$b =~ s/^(=?)(\d*)//;
				$k = $2;
				die "no value for k given" if $k eq "";
				$k = int($k);
				$arg = substr($arg,0,$i+1).$b;
				next;
			} elsif ($a eq 'S') {
				$opt_solver = 1;
				$solver_options = substr($arg,$i+1);
				$solver_options =~ s/^=//;
				$arg = substr($arg,0,$i);
				next;
			}
			die "unknown option -$a" unless defined($options->{$a});
			${$options->{$a}} = 1;
		}
	} else {
		usage() if defined($fn);
		$fn = $arg;
	}
}

usage() if $opt_help;
if ($opt_solver) {
	$solver_options .= "q" if $opt_quiet;
	$opt_quiet = 1 if $solver_options =~ /q/;
	$solver_options = "-".$solver_options if $solver_options ne "";
}
die "k must be positive" if $k == 0;

print STDERR "$PROG -- (invoke with -h for help)\n" unless $opt_quiet;
if (defined($fn)) {
	die "cannot read file $fn" unless -r $fn;
	open STDIN,$fn;
	print STDERR "reading graph from file $fn\n" unless $opt_quiet;
} else {
        print STDERR "reading graph from standard input\n" unless $opt_quiet;
}

########### read graph description ###########

$n = 0;	# nodes
@edges = ();
@loops = ();

while (<STDIN>) {
	tr/\015\012//d;
	next if (/^p ((un)?)dir/);
	next if /^c/;
	die "bad file format" unless /^\s*\d+\s*\d+\s*$/;
	($a,$b) = split;
	$n = $a if $a > $n;
	$n = $b if $b > $n;
	if ($a == $b) { push @loops,$a; next; }
	push @edges,(($a < $b)? "$a $b" : "$b $a");
}

$n++;
die "instance is trivially solvable (k >= n)" if $n <= $k;
print STDERR "generating puzzle for k=$k...\n\n" unless $opt_quiet;

########### generate the puzzle ###########

$vsp = $opt_altx? 3 : 2;
$htx = $opt_altx? 6 : 4;
$wdx = $opt_altx? 6 : 2;
@mcdev =    ('M','M','M','%');
@bigmcdev = ('M','M',' ','M');
@xdev = ('MM','MM','MM','%%');
@mcdev =    ('M','M',' ','M',' ','%') if $opt_altx;
@bigmcdev = ('M','M',' ',' ','M','%') if $opt_altx;
@xdev = ('%  M M','   M%M','%M   M','%     ',' M%M  ','%%%%%%') if $opt_altx;

# compute some important values
for ($i = 0; $i < $n-$k; $i++) { $outrow[$i] = 1+$vsp*$i; }
for ($i = $n-$k; $i < $n; $i++) { $outrow[$i] = 2+$htx*$i-$vsp*($n-$k); }
$half = $vsp*($n+$k-2)+3;
$vpos = ($n-$k)*$wdx+$k+1;
$width = $vpos+$n+2;
$height = 2*$half+1;

# puzzle with outer walls
@puzzle = ();
$baserow = 0;
for ($i = 0; $i < $height; $i++) {
	my @line = ('%',(' ')x($width-2),'%',"\n");
	push @puzzle,\@line;
}

put(0,0,'%'x$width);
put(0,$height-1,'%'x$width);

# start, restart, end
put(0,2,'S');
$dstr = $opt_diam? 'MD' : 'M';
for ($i = 1; $i < $k; $i++) { put($i,$htx*$i+2,$dstr); }
for ($i = 1; $i < $k; $i++) { put($i,$htx*($i-1)+2+$half,$dstr); }
put(0,$htx*($k-1)+2+$half,'E');

# merge/choice arrays
$pos = $k+1;
for ($i = 0; $i < $k-1; $i++) { putmc($pos,$i*$htx); }
putbigmc($pos,$i*$htx);

$pos++;
for ($i = 1; $i <= $n-$k-1; $i++) {
	$row = ($i-1)*$vsp+2;
	for ($j = 1; $j < $k; $j++) { putx($pos,$row); $row += $htx; }
	$row = $row-2+$vsp;
	putbigmc($pos,$row);
	$pos += $wdx;
}

for ($i = 0; $i < $k-1; $i++) { putmc($pos,$vsp*($n-$k)+$i*$htx); }

# horizontal wall (center left)
put($k,$half,'%'x($vpos-$k));

# central vertical wall
put($vpos,0,('%')x$height);	# vertical wall

# punch holes and put mirrors for each node
for ($i = 0; $i < $n; $i++) {
	put($vpos,$outrow[$i],' ');
	put($vpos,$outrow[$i]+$half,' ');
	put($vpos+1+$i,$outrow[$i],'M');
	put($vpos+1+$i,$outrow[$i]+$half,'M');
}

# looping edges
for $i (@loops) { put($vpos,$outrow[$i],"D"); }

# non-looping edges
for $e (@edges) {
	($a,$b) = split(' ',$e);
	put($vpos+$a+1,$outrow[$b],"D");
}

########### print puzzle to stdout or solver ###########

end:
if ($opt_solver) {
	printf "passing puzzle to solver...\n" unless $opt_quiet;
	open STDOUT,"| ./reflex.pl $solver_options";
}
for $line (@puzzle) { print join("",@$line); }
close STDOUT if $opt_solver;
exit 0;

########### auxilary procedures ###########

sub putmc {
	my ($x,$y) = @_;
	put($x,$y+1,@mcdev);
	put($x,$y+$half+1,@mcdev);
}

sub putbigmc {
	my ($x,$y) = @_;
	put($x,$y+1,@bigmcdev);
	put($x,$y+$half+1,@bigmcdev);
}

sub putx {
	my ($x,$y) = @_;
	put($x,$y+1,@xdev);
	put($x,$y+$half+1,@xdev);
}

sub put {
	my ($x,$y,@l) = @_;
	$y += $baserow;
	for my $s (@l) {
		my $line = $puzzle[$y++];
		my @nc = split(//,$s);
		splice @$line,$x,$#nc+1,@nc;
	}
}
